import os
from typing import *

import deprecation


def path_split_all(path: str) -> Tuple:
    """
    Split a given path into a list of its individual parts.

    This builds on os.path.split by doing it for all parts::

        >>> import os
        >>> os.path.split('/c/hey/there/hey.tif')
        ('/c/hey/there', 'hey.tif')
        >>> path_split_all('/c/hey/there/hey.tif')
        ('/', 'c', 'hey', 'there', 'hey.tif')
    """
    parts = []  # type: List[str]
    while True:
        split = os.path.split(path)
        if split[0] == path:
            parts.insert(0, split[0])
            break
        elif split[1] == path:
            parts.insert(0, split[1])
            break
        else:
            path = split[0]
            parts.insert(0, split[1])
    return tuple(parts)


@deprecation.deprecated("Use intelinair_utils.os_utils.list_all_files_in_dir instead")
def get_files(path: str, recursive=True, search_str=None) -> List[str]:
    """Get a list of all files within the directory"""

    def get_files_r(path, recursive):
        """Recursively processes a path returning an array of strings of file paths"""
        files = []  # type: List[str]
        path_str = os.path.join(*path)
        dir_files = os.listdir(path_str)
        dir_files.sort()
        for f in dir_files:
            local_path = os.path.join(path_str, f)
            if os.path.isfile(local_path):
                if search_str is not None:
                    if local_path.find(search_str) != -1:
                        files.append(local_path)
                else:
                    files.append(local_path)
            elif recursive and os.path.isdir(local_path):
                new_path = path + [f]
                files += get_files_r(new_path, recursive)
        return files

    path_parts = list(path_split_all(path))
    return get_files_r(path_parts, recursive)



import base64
import logging
import os
import time
from pathlib import Path
from typing import Dict, List, Tuple, Union, Optional

import boto3
import docker
import docker.errors
from docker.models.containers import Container
from requests import ReadTimeout
from tenacity import retry, wait_exponential, stop_after_attempt, retry_if_result
from intelinair_utils.logging_utils import set_standard_logging_config

set_standard_logging_config()
logger = logging.getLogger('docker_utils')


region_name = os.environ.get('AWS_DEFAULT_REGION', 'us-east-1')
docker_client = docker.from_env(timeout=300)
docker_auth = None
pulled_images = list()
retryable_errors = ('sqlite3.OperationalError',)


class ErrorLoggerException(Exception):

    def __init__(self, exit_code: int = None, message: str = None):
        super().__init__()
        self.exit_code = exit_code
        self.message = message

    def __str__(self):
        return f'ErrorLoggerException(exit_code={self.exit_code}, message={self.message})'

    def __eq__(self, other: 'ErrorLoggerException'):
        if self.exit_code == other.exit_code and self.message == other.message:
            return True
        return False


class DockerStageException(Exception):

    def __init__(self, exit_code: int = None, message: str = None):
        super().__init__()
        self.exit_code = exit_code
        self.message = message

    def __str__(self):
        return f'DockerStageException(exit_code={self.exit_code}, message={self.message})'

    def __eq__(self, other: 'DockerStageException'):
        if self.exit_code == other.exit_code and self.message == other.message:
            return True
        return False


@retry(wait=wait_exponential(multiplier=2), stop=stop_after_attempt(3), reraise=True)
def get_aws_credentials():
    """Returns frozen aws credentials given the current credentials chain"""
    session = boto3.Session()
    return session.get_credentials().get_frozen_credentials()


def get_docker_auth() -> Dict:
    """Returns the username and password for using ECR."""
    ecr = boto3.client('ecr')
    auth_creds = ecr.get_authorization_token()['authorizationData'][0]
    username, password = base64.b64decode(auth_creds['authorizationToken']).decode('utf-8').split(':')
    return {
        'username': username,
        'password': password
    }


def remove_old_docker_images(repo: str, tag: str):
    """
    Remove old images of the same repo if exist
    Args:
        repo: image repository name
        tag: image tag
    """
    docker_images = docker_client.images.list(name=repo)
    for docker_image in docker_images:
        for _tag in docker_image.tags:
            if not _tag.endswith(tag):
                logger.info(f"Removing older image version {_tag}")
                docker_client.images.remove(docker_image.id, force=True)


def _log_pull_generator(pull_logs: List, log_period_seconds: int = 30):
    """Generator for pull logs"""
    t = time.time()
    last_log = ""
    for log in pull_logs:
        last_log = log
        if time.time() - t > log_period_seconds:
            t = time.time()
            logger.info(last_log)
    if last_log:
        logger.info(last_log)


def pull_docker_image(image: str) -> Union[None, str]:
    """Ensures the image is pulled locally

    Args:
        image: the docker image to pull

    Returns:
        an error message upon failure otherwise None
    """
    global docker_auth, pulled_images
    try:
        if ':' in image:
            repo, tag = image.split(':')
        else:
            repo, tag = image, 'latest'
    except Exception as e:
        logger.error('Failed to parse docker image tag')
        logger.exception(e, exc_info=True)
        raise e

    if (repo, tag) in pulled_images:
        return

    if os.environ.get('IN_EC2') == 'YES':
        for _ in range(3):
            try:
                remove_old_docker_images(repo, tag)
            except docker.errors.APIError:
                logger.error("Failed to remove older image version. Retrying")
                time.sleep(30)
            else:
                break
        else:
            logger.error(f"Failed to remove older image versions for {repo}")

    try:
        logger.info(f'Pulling docker image = {repo}:{tag}')
        pull_logs = docker_client.api.pull(repo, tag=tag, auth_config=docker_auth, stream=True, decode=True)
        _log_pull_generator(pull_logs)
        logger.info("Docker is successfully pulled")
    except docker.errors.APIError:
        logger.info('Failed to pull image, attempting to get new credentials')
        docker_auth = get_docker_auth()
        logger.info('Successfully got credentials')
        try:
            pull_logs = docker_client.api.pull(repo, tag=tag, auth_config=docker_auth, stream=True, decode=True)
            _log_pull_generator(pull_logs)
            logger.info("Docker is successfully pulled")
        except docker.errors.APIError as e:
            logger.error('Failed to pull docker image')
            logger.exception(e, exc_info=True)
            raise e
    finally:
        pulled_images.append((repo, tag))


def check_source_volumes_exist(volumes: Optional[List[str]]):
    """Function that checks whether given list of volumes exist."""
    if not volumes:
        return
    for v in volumes:
        source_volume, _ = v.split(':')
        source_path = Path(source_volume).expanduser()
        if not source_path.exists():
            logger.info(f'Creating source volume directory {source_path}')
            source_path.mkdir(parents=True, mode=0o777)


def is_retryable_return_value(exception):
    if isinstance(exception, DockerStageException):
        for retryable_error in retryable_errors:
            if retryable_error in exception.message:
                return True
        raise exception
    return False


@retry(
    retry=retry_if_result(is_retryable_return_value)
)
def call_docker(
        image: str,
        python_script_path: str,
        command_name: str = '',
        runtime: str = None,
        script_args: list = None,
        script_kwargs: dict = None,
        shared_memory: str = None,
        user: str = 'ubuntu',
        group: str = 'ubuntu',
        working_dir: str = None,
        volumes: List[str] = None,
        environment: Dict = None,
        max_runtime: int = 60 * 5,
        pipeline_version: Optional[str] = None,
) -> None:
    """Starts the script in the image with the specified kwargs

    Args:
        image: The docker image to pull and use
        python_script_path: the python script to run inside the image
        command_name: docker command name
        runtime: the container runtime to use
        script_args: a list of args to pass to the script
        script_kwargs: the kwargs to pass to the script, will pass like --{key} {value} for each item
        shared_memory: the amount of shared memory to pass to the container
        user: the user to run the docker container as
        group: the group to run the docker container in
        working_dir: the directory to use as the working dir inside the container
        volumes: A list of volumes to mount
        environment: a dictionary of extra environment variables for the container
        max_runtime: the max number of seconds to wait for the container to finish
        pipeline_version: the version of the pipeline to run the container if function is used in image processing.

    Returns:
        an error message upon failure or None otherwise
    """
    ret_code = None
    error_msg = f'Failed to process docker command {command_name}'

    try:
        pull_docker_image(image=image)
        check_source_volumes_exist(volumes)
        ret_code, error_msg, from_error_logger = run_container(
            image=image,
            script_args=script_args,
            script_kwargs=script_kwargs,
            python_script_path=python_script_path,
            runtime=runtime,
            volumes=volumes,
            shared_memory=shared_memory,
            user=user,
            group=group,
            working_dir=working_dir,
            environment=environment,
            max_runtime=max_runtime,
            pipeline_version=pipeline_version
        )

    except (TimeoutError, MemoryError, ReadTimeout) as e:
        logger.warning("Attempting a rerun on a larger machine.")
        logger.warning(e)
        logger.warning(e.__class__)
        raise MemoryError()

    except Exception as e:
        logger.error(error_msg)
        logger.exception(e)
        if ret_code is not None and ret_code != 0:
            error_msg = f"{error_msg}, Return code: {ret_code}"
        raise Exception(error_msg)
    else:
        if ret_code != 0 and from_error_logger:
            raise ErrorLoggerException(exit_code=ret_code, message=error_msg)
        elif ret_code != 0:
            raise DockerStageException(exit_code=ret_code, message=error_msg)


# Debug utility
def build_docker_command(image: str,
                         command: str,
                         shm_size: str,
                         user: str,
                         volumes: List,
                         environment: Dict,
                         **kwargs: Dict):
    """Builds a docker command from the given parameters"""
    docker_command = f'docker run --rm --net host --shm-size={shm_size} --user {user} '
    docker_command += ' '.join([f'-v {v}' for v in volumes]) + ' '
    docker_command += ' '.join(f'--env "{k}={environment[k]}" ' for k in environment) + ' '
    docker_command += f' {image} ' + ' '.join(command)
    return docker_command


def run_container(
        image: str,
        user: str,
        group: str,
        python_script_path: str,
        runtime: str,
        script_args: List,
        script_kwargs: Dict,
        shared_memory: str,
        working_dir: str,
        volumes: List[str],
        environment: Dict,
        max_runtime: int,
        pipeline_version: Optional[str] = None,
) -> Tuple:
    """Starts the script in the image with the specified kwargs

    Args:
        image: The docker image to pull and use
        python_script_path: the python script to run inside the image
        runtime: the docker runtime to execute with (eg nvidia)
        script_args: a list of args to pass to the script
        script_kwargs: the kwargs to pass to the script, will pass like --{key} {value} for each item
        shared_memory: the amount of shared memory to pass to the container
        user: the user to run the docker container as
        group: the group to run the docker container as
        working_dir: the directory to use as the working directory inside the container
        volumes: A list of volumes to mount
        environment: a dictionary of extra environment variables for the container
        max_runtime: the max number of seconds to wait for the container to finish
        pipeline_version: the version of the pipeline to run the container if function is used in image processing.

    Returns:
        an error message upon failure or None otherwise
    """

    credentials = get_aws_credentials()

    final_script_args = list()

    if script_args:
        final_script_args.extend(script_args)

    if script_kwargs:
        for k in script_kwargs:
            final_script_args.extend([f'--{k}', script_kwargs[k]])

    run_kwargs = {
        'image': image,
        'command': [str(a) for a in ['python3', python_script_path] + final_script_args],
        'runtime': runtime,
        'shm_size': shared_memory,
        'user': user,
        'group_add': [group],
        'network': 'host',
        'volumes': volumes,
        'working_dir': working_dir,
        'environment': {
            'AWS_DEFAULT_REGION': region_name,
            'AWS_ACCESS_KEY_ID': credentials.access_key,
            'AWS_SECRET_ACCESS_KEY': credentials.secret_key,
            'AWS_SESSION_TOKEN': credentials.token if credentials.token else '',
            # set ENV variable  IA_LOGGING_FMT
            # getting root logger's handler formatter
            'IA_LOGGING_FMT': logging.getLogger().handlers[0].formatter._fmt,
            'IA_DOCKER_IMAGE': image,
        },
        'detach': True
    }
    if pipeline_version is not None:
        run_kwargs['environment']['IA_PIPELINE_VERSION'] = pipeline_version

    if environment:
        run_kwargs['environment'].update(environment)

    # set a breakpoint here to get a docker command you can test in isolation
    # debug_command = self.build_docker_command(**run_kwargs)
    container = docker_client.containers.run(**run_kwargs)
    return log_and_shutdown_container(container=container, max_runtime=max_runtime,
                                      output_dir=script_kwargs.get('output_dir'))


def log_and_shutdown_container(container: Container, max_runtime: int, output_dir: str = None) -> Tuple[int, str, bool]:
    """Logs the container and shut it down."""
    logs = []
    for attempt in range(max_runtime):
        time.sleep(1)
        container.reload()
        logs = container.logs(stream=True, follow=False)
        if not container.attrs['State']['Running']:
            break

    else:
        logger.error(f"The container with image {container.image} could "
                       f"not finish the job in {max_runtime} seconds >> LOGS:")

        for log in logs:
            decoded_log = log.decode('utf-8').strip('\n')
            logger.warning(decoded_log, extra={'simple': True})
        raise TimeoutError()

    decoded_logs = []
    logs = dict.fromkeys(logs)
    for log in logs:
        # if logging receives 'simple' kwarg,
        # it disables current format and logs as is.

        decoded_log = log.decode('utf-8').strip('\n')
        logger.info(decoded_log, extra={'simple': True})
        decoded_logs.append(decoded_log)

    logger.info(f'Container status: {container.status}')
    exit_result = container.wait()
    ret_code = exit_result['StatusCode']
    logger.info(f'Container exit code: {ret_code}')

    container.remove()
    error_message = None
    error_file_exists = os.path.exists(os.path.join(output_dir, 'error.txt')) if output_dir else False
    if ret_code in (137, 247):  # Killed for using too much memory
        logger.warning(f"Container returned exit code {ret_code}")
        raise MemoryError()
    elif ret_code != 0 and error_file_exists:
        try:
            with open(os.path.join(output_dir, 'error.txt'), 'r') as fp:
                error_message = fp.read().split('\n')[1].strip()
        except Exception as e:
            logger.error(repr(e))
            logger.warning(f"Could not retrieve error message from error logger.")

    elif ret_code != 0 and not error_file_exists:
        # In case of failure the last block of errors logged streamed from inside
        # the container is printed along with the stats log for easier debugging.
        try:
            error_message = get_last_traceback(decoded_logs=decoded_logs)
        except Exception as e:
            logger.error(repr(e))
            logger.warning(f"Could not retrieve the last logs of the job, that failed inside the container.")

    return ret_code, error_message, error_file_exists


def get_last_traceback(decoded_logs: List[str]) -> str:
    """
    Args:
        decoded_logs: The list of logs streamed in docker

    Returns:
        EITHER The last block of logs beginning from 'Traceback' formatted
        in the blank OR the full logs (when no 'Traceback' line was found).
    """
    inverted_error_logs = decoded_logs[::-1]  # starting from end
    is_short = lambda ll, i: len(ll) < i + 1  # recursion safe exit
    is_traceback = lambda elem: 'Traceback' in elem

    # getting the last block
    get_last = lambda ll, i: ll[:i + 1] if is_short(ll, i) or is_traceback(ll[i]) else get_last(ll, i + 1)

    traceback_block_as_list = get_last(inverted_error_logs, 0)
    traceback_block = "\n".join(traceback_block_as_list[::-1])  # re-inverted

    return f"{traceback_block}"



"""Planting service API wrapper class."""

import json
from typing import Union, Optional

from shapely import wkt
import shapely.geometry
import shapely.wkt
from pydantic import BaseModel, field_validator, ConfigDict

from intelinair_utils.api_wrapper import ApiWrapper



class Geometry(BaseModel):
    """Model for geometry data supporting multiple formats."""
    geometry: Optional[Union[shapely.geometry.MultiPolygon, shapely.geometry.Polygon, dict, str]]

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        json_encoders={
            shapely.geometry.MultiPolygon: lambda s: shapely.geometry.mapping(s) if s else None,
            shapely.geometry.Polygon: lambda s: shapely.geometry.mapping(s) if s else None,
        }
    )

    @field_validator('geometry', mode='before')
    @classmethod
    def parse_geometry(cls, value):
        """Validate geometry as either WKT, GeoJSON, or a shapely object."""
        return cls.validate_geometry(value)

    def dict_jsonable(self):
        """Return dictionary which is serializable to json."""
        return json.loads(self.model_dump_json())

    @classmethod
    def validate_geometry(cls, value):
        """Validates that the geometry is a valid Polygon, MultiPolygon, WKT, or GeoJSON."""
        if value:
            if isinstance(value, (shapely.geometry.MultiPolygon, shapely.geometry.Polygon)):
                return value
            elif isinstance(value, str):
                try:
                    shapely.wkt.loads(value)  # Validate as WKT string
                except shapely.errors.WKTReadingError:
                    raise ValueError(f"Invalid WKT string: {value}")
            elif isinstance(value, dict):
                try:
                    shapely.geometry.shape(value)  # Validate as GeoJSON dictionary
                except (TypeError, ValueError):
                    raise ValueError(f"Invalid GeoJSON dictionary: {value}")
            else:
                raise ValueError(f"Unsupported geometry type: {type(value).__name__}")
        return value


class ElevationTile(Geometry):
    """Model for elevation data."""
    name: str
    s3_path: str
    north_east_latitude: float
    north_east_latitude: float
    north_east_longitude: float
    south_west_latitude: float
    south_west_longitude: float
    resolution: float
    country: str
    geometry: Optional[Union[shapely.geometry.MultiPolygon, shapely.geometry.Polygon, dict, str]]

    @field_validator('geometry', mode='before')
    @classmethod
    def parse_geometry(cls, value):
        """Validate geometry as either WKT, GeoJSON, or a shapely object."""
        return cls.validate_geometry(value)


class ElevationMetadataServiceApi(ApiWrapper):
    """API wrapper for interacting with the Elevation Metadata Service."""

    ELEVATION_METADATA_SERVICE_URLS = {
        "prod": "https://utl43kq9za.execute-api.us-east-1.amazonaws.com/api",
        "release": "https://p8u69ef8d4.execute-api.us-east-1.amazonaws.com/api",
        "longview": "https://i7upioahk4.execute-api.us-east-1.amazonaws.com/api",
    }

    def __init__(self, environment: str, config_path: Optional[str] = None):
        super().__init__(environment, self._get_api_urls(), config_path=config_path)

    def _get_api_urls(self):
        """Overridden method for getting elevation metadata service API gateway URLs."""
        return self.ELEVATION_METADATA_SERVICE_URLS

    def create_tile(self, elevation_tile: ElevationTile):
        """Create a new elevation tile and return the created ElevationTile object."""
        url = self.url_template.format("create-tile-info")
        response = self._make_request(self.session.post, url, json=elevation_tile.dict_jsonable())
        response.raise_for_status()
        return ElevationTile(**response.json())

    def get_tiles_s3_paths_by_lonlat(self, ne_lat, ne_lon, sw_lat, sw_lon):
        """Retrieve S3 paths of tiles intersecting the specified latitude and longitude bounds."""
        url = self.url_template.format(f"/get-intersected-tiles-by-lonlat/{ne_lat}/{ne_lon}/{sw_lat}/{sw_lon}")
        response = self._make_request(self.session.get, url)
        response.raise_for_status()
        return response.json()

    def get_tiles_s3_paths_by_geom(self, geometry: Geometry):
        """Retrieve S3 paths of tiles intersecting the specified geometry."""
        url = self.url_template.format("get-intersected-tiles-by-geom")
        response = self._make_request(self.session.post, url, json=geometry.dict_jsonable())
        response.raise_for_status()
        return response.json()

    def get_tile_by_name(self, tile_name: str):
        """Retrieve tile info by name."""
        url = self.url_template.format(f"get-tile-info/{tile_name}")
        response = self._make_request(self.session.get, url)
        response.raise_for_status()
        return response.json()



import json
import logging
import os
from collections import defaultdict
from datetime import datetime
from typing import Generator, List, Optional, Tuple, Dict

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from intelinair_utils.api import Api

logger = logging.getLogger(__name__)

KNOWN_ENVS = ['prod', 'release', 'bellflower', 'local']
DEFAULT_FILTER_PATH = os.path.expanduser('~/.mlapi_filters.json')
FILTER_PATH_ENV_KEY = 'MLAPI_FILTER_PATH'


class MLApi(Api):
    """
    Wrapper class for queries against the ML Service

    Supports a config file which can be used to filter information returned by this client,
    The default config location is '~/.mlapi_filters.json' and can be provided by setting
    the MLAPI_FILTER_PATH environment variable

    The config file should be a json file with the following keys:
        `inference_filter_list`    : A dictionary of filter params to include for the get_inferences function
                                     The parameter should be a dictionary with the key being an inference_type
                                     and the value should be extra filter parameters to include with the request
                                     to the ML Api.
                                     Example:
                                         {
                                            "inference_type1": {"model_name": "your-model", "model_version": 2},
                                            "inference_type2": {"model_id": 2, "docker_image": "model_image:tag"}
                                        }

        `inference_type_allow_list`: A list of allowed inference types for the get_inferences function to return
                                     The function will raise a ValueError for any request for an inference_type
                                     not in this allow list
                                     Example:
                                         ["inference_type1", "inference_type2"]

    """

    def __init__(self, environment):
        # version_string == '' corresponds to the first version of API
        assert environment in KNOWN_ENVS, 'Unknown environment for the ML Service'
        self.environment = environment
        if environment == 'local':
            self.api_url = 'http://localhost:8080'
        else:
            self.api_url = f'https://ml.{environment}.int.intelinair.dev'
        # all calls (other than login) should use the version string
        self.session = requests.Session()
        retries = Retry(total=10, read=5, connect=5, backoff_factor=1, allowed_methods=False,
                        status_forcelist=[104, 500, 502, 503, 504])
        self.session.mount('https://', HTTPAdapter(max_retries=retries))
        self.session.mount('http://', HTTPAdapter(max_retries=retries))

        filter_path = os.environ.get(FILTER_PATH_ENV_KEY, DEFAULT_FILTER_PATH)
        filter_content = self._load_filters(filter_path)
        self.inference_filters = filter_content.get('inference_filter_list', {})
        self.inference_type_allow_list = filter_content.get('inference_type_allow_list', None)

    @staticmethod
    def _load_filters(filter_path):
        """Loads the filter config"""
        if os.path.exists(filter_path):
            with open(filter_path) as fp:
                return json.load(fp)
        else:
            return {}

    def get(self, path, params=None, json=None, ignore_json=False):
        res = self.session.get(self.api_url + path, params=params, json=json)
        res.raise_for_status()
        if ignore_json is True:
            return res
        else:
            return res.json()

    def post(self, path, json=None, params=None):
        res = self.session.post(self.api_url + path, json=json, params=params)
        res.raise_for_status()
        return res.json()

    def patch(self, path, json=None, params=None):
        res = self.session.patch(self.api_url + path, json=json, params=params)
        res.raise_for_status()
        return res.json()

    def delete(self, path, json=None, params=None):
        res = self.session.delete(self.api_url + path, json=json, params=params)
        res.raise_for_status()
        return res.json()

    def create_dataset(self, name: str, version: int, data: dict, tags: dict = None) -> dict:
        """Creates a dataset with the given parameters"""
        return self.post('/datasets/', json={
            'name': name,
            'version': version,
            'data': data,
            'tags': tags
        })

    def update_dataset(self, dataset: dict) -> dict:
        """Updates a dataset object in the service with new data"""
        return self.patch(f'/datasets/{dataset["id"]}', json=dataset)

    def get_dataset_by_id(self, dataset_id: int) -> dict:
        """Gets a dataset by its id"""
        return self.get(f'/datasets/{dataset_id}')

    def get_latest_dataset(self, dataset_name: str) -> dict:
        """Gets the latest dataset with the given name"""
        return self.get(f'/datasets/latest', params={'dataset_name': dataset_name})

    def get_datasets(self, dataset_name: str = None, dataset_version: int = None,
                     api_limit_size: int = 100) -> Generator[dict, None, None]:
        """Gets a generator of datasets optionally filtering by name and/or version"""
        skip = 0
        while True:
            datasets = self.get(
                '/datasets/',
                params={
                    'skip': skip,
                    'limit': api_limit_size,
                    'name': dataset_name,
                    'version': dataset_version
                }
            )
            for dataset in datasets:
                yield dataset
            if len(datasets) == api_limit_size:
                skip += api_limit_size
            else:
                break

    def get_dataset(self, dataset_name: str, dataset_version: int = None) -> dict:
        """Gets a dataset by name and optionally a specific version of the dataset"""
        if dataset_version is None:
            return self.get_latest_dataset(dataset_name)
        datasets = list(self.get_datasets(dataset_name, dataset_version))
        if len(datasets) == 0:
            raise Exception(f'Could not find dataset with  '
                            f'dataset_name={dataset_name} and version={dataset_version}')
        try:
            [dataset] = datasets
            return dataset
        except ValueError:
            raise ValueError(f'Multiple datasets registered with the same version '
                             f'model_name={dataset_name} and version={dataset_version}')

    def get_dataset_flights(self, dataset_id: int, api_limit_size: int = 250) -> Dict[str, List[dict]]:
        """Gets all the flights in a dataset"""
        offset = 0
        final_results = defaultdict(list)
        while True:
            results = self.get(f'/datasets/flights',
                               json={
                                   'dataset_id': dataset_id,
                                   'offset': offset,
                                   'limit': api_limit_size,
                               })
            for split, flights in results.items():
                final_results[split].extend(flights)
            if sum([len(flights) for flights in results.values()]) == api_limit_size:
                offset += api_limit_size
            else:
                break
        return final_results

    def create_model(self, model_name: str, model_version: int, model_hash: str, s3_location: str,
                     tags: dict = None, data: dict = None) -> dict:
        """Creates a model with the ML Service with the given parameters"""
        return self.post(
            '/models/',
            json={
                "model_name": model_name,
                "model_version": model_version,
                "model_hash": model_hash,
                "s3_location": s3_location,
                "tags": tags,
                "data": data
            }
        )

    def update_model(self, model: dict) -> dict:
        """Updates a model"""
        return self.patch(f'/models/{model["id"]}', json=model)

    def get_or_create_model_from_script(self, git_repo_name: str, script_name: str, git_description: str) -> dict:
        """Gets a model by script name and git description or creates one if it doesn't exist"""
        return self.post('/models/script', params={
            'git_repo_name': git_repo_name,
            'script_name': script_name,
            'git_description': git_description
        })

    def get_model_by_id(self, model_id: int) -> dict:
        """Gets a model by id"""
        return self.get(f'/models/{model_id}')

    def get_latest_model(self, model_name: str) -> dict:
        """Gets the latest model with the given name"""
        return self.get(f'/models/latest', params={'model_name': model_name})

    def get_models(self, model_name: str = None, model_version: int = None,
                   api_limit_size: int = 100) -> Generator[dict, None, None]:
        """Gets models from the api with optional filters"""
        skip = 0
        while True:
            models = self.get(f'/models/?skip={skip}&limit={api_limit_size}',
                              json={'model_name': model_name, 'model_version': model_version})
            for model in models:
                yield model
            if len(models) == api_limit_size:
                skip += api_limit_size
            else:
                break

    def get_model(self, model_name: str, model_version: int = None) -> dict:
        """Gets the model with the give name and version, if no version is specified the latest version is returned"""
        if model_version is None:
            return self.get_latest_model(model_name)
        models = list(self.get_models(model_name, model_version))
        if len(models) == 0:
            raise Exception(f'Could not find registered models with '
                            f'model_name={model_name} and version={model_version}')
        try:
            [model] = models
            return model
        except ValueError:
            raise ValueError(f'Multiple models registered with the same version '
                             f'model_name={model_name} and version={model_version}')

    def create_inference(self, flight_code: str, inference_type: str, model_id: int, docker_image: str,
                         hparams: dict, result: dict, zone_token: Optional[str] = None,
                         pipeline_version: Optional[str] = None) -> dict:
        """Creates an inference with the given parameters"""
        return self.post('/inferences/flight/', json={
            'flight_code': flight_code,
            'zone_token': zone_token,
            'inference_type': inference_type,
            'model_id': model_id,
            'docker_image': docker_image,
            'pipeline_version': pipeline_version,
            'hparams': hparams,
            'result': result,
        })

    def create_inference_from_script(self, flight_code: str, inference_type: str, result: dict, git_repo_name: str,
                                     script_name: str, git_description: str, hparams: dict,
                                     zone_token: Optional[str] = None, docker_image: Optional[str] = None,
                                     pipeline_version: Optional[str] = None) -> dict:
        """Creates an inference from a script"""
        return self.post('/inferences/flight/script', json={
            'flight_code': flight_code,
            'zone_token': zone_token,
            'inference_type': inference_type,
            'git_repo_name': git_repo_name,
            'script_name': script_name,
            'git_description': git_description,
            'docker_image': docker_image,
            'pipeline_version': pipeline_version,
            'hparams': hparams,
            'result': result
        })

    def update_inference(self, inference: dict) -> dict:
        return self.patch(f'/inferences/flight/{inference["id"]}', json=inference)

    def get_inference_by_id(self, inference_id: int) -> dict:
        """Gets an inference by id"""
        return self.get(f'/inferences/flight/{inference_id}')

    def get_inferences(self, flight_code: List[str] = None, inference_type: str = None, model_id: int = None,
                       model_name: str = None, model_version: int = None, git_repo_name: str = None,
                       script_name: str = None, git_description: str = None, docker_image: str = None,
                       pipeline_version: str = None, hparams: dict = None,
                       start_ts: datetime = None, end_ts: datetime = None,
                       zone_token: str = None, api_limit_size: int = 100) -> Generator[dict, None, None]:
        """Returns inferences that match the provided filters"""
        if self.inference_type_allow_list and inference_type not in self.inference_type_allow_list:
            raise ValueError(f"inference type: {inference_type} not in "
                             f"inference type allow list {self.inference_type_allow_list}")

        request = {
            'flight_code': flight_code,
            'inference_type': inference_type,
            'zone_token': zone_token,
            'start_ts': str(start_ts) if start_ts else None,
            'end_ts': str(end_ts) if end_ts else None,
            'model_id': model_id,
            'model_name': model_name,
            'model_version': model_version,
            'git_repo_name': git_repo_name,
            'script_name': script_name,
            'git_description': git_description,
            'docker_image': docker_image,
            'pipeline_version': pipeline_version,
            'hparams': hparams
        }

        if inference_type and inference_type in self.inference_filters:
            request.update(**self.inference_filters[inference_type])

        base_id = 0
        while True:
            response = self.get(f'/v2/inferences/flight/?base_id={base_id}&limit={api_limit_size}', json=request)
            for inference in response['inferences']:
                yield inference
            if response['next_id'] is not None:
                base_id = response['next_id']
            else:
                break

    def get_inference_types(self, flight_code: str) -> List[str]:
        """Returns the inference types available for this flight"""
        return self.get(f'/inferences/flight/types/{flight_code}')

    def get_flight_codes_with_inference_type(self, inference_type: str,
                                             api_limit_size: int = 10000) -> Generator[str, None, None]:
        """Returns a list of flight codes with an inference with the given inference type"""
        skip = 0
        while True:
            flight_codes = self.get(f'/inferences/flight/codes/{inference_type}?skip={skip}&limit={api_limit_size}')
            for flight_code in flight_codes:
                yield flight_code
            if len(flight_codes) == api_limit_size:
                skip += api_limit_size
            else:
                break

    def create_flight_evaluation(self, flight_code: str, inference_id: int, scores: dict, eval_type: str = 'default') -> dict:
        """Creates an evaluation for a flight with the given parameters"""
        return self.post('/evaluations/flight/', json={
            'flight_code': flight_code,
            'inference_id': inference_id,
            'type': eval_type,
            'scores': scores,
        })

    def get_flight_evaluation_by_id(self, flight_evaluation_id: int) -> dict:
        """Returns an evaluation by id"""
        return self.get(f'/evaluations/flight/{flight_evaluation_id}')

    def get_flight_evaluations(self, flight_code: List[str] = None, inference_type: str = None,
                               model_id: int = None, model_name: str = None, model_version: int = None,
                               git_repo_name: str = None, script_name: str = None, git_description: str = None,
                               docker_image: str = None, hparams: dict = None,
                               inference_id: int = None, eval_type: str = None,
                               api_limit_size: int = 100) -> Generator[dict, None, None]:
        """Generates a set of evaluations that match the provided filters"""
        skip = 0
        while True:
            evaluations = self.get(
                '/evaluations/flight/',
                json={
                    'flight_code': flight_code,
                    'inference_type': inference_type,
                    'type': eval_type,
                    'model_id': model_id,
                    'model_name': model_name,
                    'model_version': model_version,
                    'git_repo_name': git_repo_name,
                    'script_name': script_name,
                    'git_description': git_description,
                    'docker_image': docker_image,
                    'hparams': hparams,
                    'inference_id': inference_id,
                },
                params={
                    'skip': skip,
                    'limit': api_limit_size,
                }
            )
            for evaluation in evaluations:
                yield evaluation
            if len(evaluations) == api_limit_size:
                skip += api_limit_size
            else:
                break

    def create_field_inference(self, field_token: str, date_active: datetime, inference_type: str, model_id: int,
                               docker_image: str, hparams: dict, result: dict, season_id: int = -1,
                               zone_token: str = 'boundary', pipeline_version: Optional[str] = None) -> dict:
        """Creates an inference with the given parameters"""
        return self.post('/inferences/field/', json={
            'field_token': field_token,
            'season_id': season_id,
            'zone_token': zone_token,
            'date_active': date_active.isoformat(),
            'inference_type': inference_type,
            'model_id': model_id,
            'docker_image': docker_image,
            'pipeline_version': pipeline_version,
            'hparams': hparams,
            'result': result,
        })

    def create_field_inference_from_script(self, field_token: str, date_active: datetime, inference_type: str,
                                           result: dict,
                                           git_repo_name: str, script_name: str, git_description: str, hparams: dict,
                                           season_id: int = -1, zone_token: str = 'boundary',
                                           docker_image: Optional[str] = None,
                                           pipeline_version: Optional[str] = None) -> dict:
        """Creates an inference from a script"""
        return self.post('/inferences/field/script', json={
            'field_token': field_token,
            'season_id': season_id,
            'zone_token': zone_token,
            'date_active': date_active.isoformat(),
            'inference_type': inference_type,
            'git_repo_name': git_repo_name,
            'script_name': script_name,
            'git_description': git_description,
            'docker_image': docker_image,
            'pipeline_version': pipeline_version,
            'hparams': hparams,
            'result': result
        })

    def update_field_inference(self, inference: dict) -> dict:
        return self.patch(f'/inferences/field/{inference["id"]}', json=inference)

    def get_field_inference_by_id(self, inference_id: int) -> dict:
        """Gets an inference by id"""
        return self.get(f'/inferences/field/{inference_id}')

    def get_field_inferences(self, date: datetime = datetime.max, field_token: List[str] = None,
                             season_id: List[int] = None,
                             zone_token: List[str] = None, inference_type: str = None, model_id: int = None,
                             model_name: str = None,
                             model_version: int = None, git_repo_name: str = None, script_name: str = None,
                             git_description: str = None, docker_image: str = None, pipeline_version: str = None,
                             hparams: dict = None, start_ts: datetime = None, end_ts: datetime = None,
                             api_limit_size: int = 100) -> Generator[dict, None, None]:
        """Returns inferences that match the provided filters"""
        if self.inference_type_allow_list and inference_type not in self.inference_type_allow_list:
            raise ValueError(f"inference type: {inference_type} not in "
                             f"inference type allow list {self.inference_type_allow_list}")

        request = {
            'date': date.isoformat(),
            'field_token': field_token,
            'season_id': season_id,
            'inference_type': inference_type,
            'zone_token': zone_token,
            'start_ts': str(start_ts) if start_ts else None,
            'end_ts': str(end_ts) if end_ts else None,
            'model_id': model_id,
            'model_name': model_name,
            'model_version': model_version,
            'git_repo_name': git_repo_name,
            'script_name': script_name,
            'git_description': git_description,
            'docker_image': docker_image,
            'pipeline_version': pipeline_version,
            'hparams': hparams
        }

        if inference_type and inference_type in self.inference_filters:
            request.update(**self.inference_filters[inference_type])

        base_id = 0
        while True:
            response = self.get(f'/inferences/field/?base_id={base_id}&limit={api_limit_size}', json=request)
            for inference in response['inferences']:
                yield inference
            if response['next_id'] is not None:
                base_id = response['next_id']
            else:
                break

    def get_latest_field_inferences(
        self, field_token: List[str], date: datetime = datetime.max, season_id: List[int] = None,
        zone_token: List[str] = None,
        inference_type: str = None, model_id: int = None, model_name: str = None,
        model_version: int = None, git_repo_name: str = None, script_name: str = None,
        git_description: str = None, docker_image: str = None, pipeline_version: str = None,
        hparams: dict = None, start_ts: datetime = None, end_ts: datetime = None, api_limit_size: int = 100
    ) -> Generator[Tuple[Tuple[str, int], dict], None, None]:
        """Returns the latest field inferences that match the given filters
        After the filters are applied inferences are grouped by field_toke, season_id, zone_token, and inference type
        For each of these groups the latest inference is defined to be in the group of inferences with the latest active date
        and among these the inference with the highest version is selected

        Yields:
            key: the key for the latest inference (field_token, season_id, zone_token, inference_type)
            inference: the dict of the inference
        """
        if self.inference_type_allow_list and inference_type not in self.inference_type_allow_list:
            raise ValueError(f"inference type: {inference_type} not in "
                             f"inference type allow list {self.inference_type_allow_list}")

        request = {
            'date': date.isoformat(),
            'season_id': season_id,
            'inference_type': inference_type,
            'zone_token': zone_token,
            'start_ts': str(start_ts) if start_ts else None,
            'end_ts': str(end_ts) if end_ts else None,
            'model_id': model_id,
            'model_name': model_name,
            'model_version': model_version,
            'git_repo_name': git_repo_name,
            'script_name': script_name,
            'git_description': git_description,
            'docker_image': docker_image,
            'pipeline_version': pipeline_version,
            'hparams': hparams
        }

        if inference_type and inference_type in self.inference_filters:
            request.update(**self.inference_filters[inference_type])

        for i in range(0, len(field_token), api_limit_size):
            request['field_token'] = field_token[i:i + api_limit_size]
            response: List = self.get(f'/inferences/field/latest', json=request)
            for inference in response:
                key = (
                    inference['field_token'], inference['season_id'],
                    inference['zone_token'], inference['inference_type']
                )
                yield key, inference

    def get_field_inference_types(self, field_token: str) -> List[str]:
        """Returns the inference types available for this field"""
        return self.get(f'/inferences/field/types/{field_token}')

    def get_field_tokens_with_inference_type(self, inference_type: str,
                                             api_limit_size: int = 10000) -> Generator[str, None, None]:
        """Returns a list of field tokens with an inference with the given inference type"""
        skip = 0
        while True:
            field_tokens = self.get(f'/inferences/field/tokens/{inference_type}?skip={skip}&limit={api_limit_size}')
            for field_token in field_tokens:
                yield field_token
            if len(field_tokens) == api_limit_size:
                skip += api_limit_size
            else:
                break

    def create_field_evaluation(self, inference_id: int, scores: dict, eval_type: str = 'default') -> dict:
        """Creates an evaluation for a field with the given parameters"""
        return self.post('/evaluations/field/', json={
            'inference_id': inference_id,
            'type': eval_type,
            'scores': scores,
        })

    def get_field_evaluation_by_id(self, field_evaluation_id: int) -> dict:
        """Returns an evaluation by id"""
        return self.get(f'/evaluations/field/{field_evaluation_id}')

    def get_field_evaluations(self, date: datetime = datetime.max, field_token: List[str] = None,
                              season_id: List[int] = None,
                              zone_token: List[str] = None, inference_type: str = None, model_id: int = None,
                              model_name: str = None, model_version: int = None, git_repo_name: str = None,
                              script_name: str = None, git_description: str = None, docker_image: str = None,
                              hparams: dict = None, inference_id: List[int] = None, eval_type: str = None,
                              api_limit_size: int = 100) -> Generator[dict, None, None]:
        """Generates a set of evaluations that match the provided filters"""
        skip = 0
        while True:
            evaluations = self.get(
                '/evaluations/field/',
                json={
                    'date': date.isoformat(),
                    'field_token': field_token,
                    'season_id': season_id,
                    'zone_token': zone_token,
                    'inference_type': inference_type,
                    'model_id': model_id,
                    'model_name': model_name,
                    'model_version': model_version,
                    'git_repo_name': git_repo_name,
                    'script_name': script_name,
                    'git_description': git_description,
                    'docker_image': docker_image,
                    'hparams': hparams,
                    'inference_id': inference_id,
                    'type': eval_type,
                },
                params={
                    'skip': skip,
                    'limit': api_limit_size,
                }
            )
            for evaluation in evaluations:
                yield evaluation
            if len(evaluations) == api_limit_size:
                skip += api_limit_size
            else:
                break

    def create_dataset_evaluation(self, dataset_id: int, model_id: int, scores: dict, hparams: dict, tags: dict = None,
                                  docker_image: str = None, pipeline_version: str = None) -> dict:
        """Creates an evaluation for the given dataset and model"""
        return self.post('/evaluations/dataset/', json={
            'dataset_id': dataset_id,
            'model_id': model_id,
            'scores': scores,
            'hparams': hparams,
            'tags': tags,
            'docker_image': docker_image,
            'pipeline_version': pipeline_version
        })

    def get_dataset_evaluation_by_id(self, dataset_evaluation_id: int) -> dict:
        """Returns a dataset evaluation by id"""
        return self.get(f'/evaluations/dataset/{dataset_evaluation_id}')

    def get_dataset_evaluations(self, model_id: id = None, docker_image: str = None, hparams: dict = None,
                                dataset_id: int = None, pipeline_version: str = None,
                                api_limit_size: int = 100) -> Generator[dict, None, None]:
        """Generates a set of dataset evaluations that match the given parameters"""
        skip = 0
        while True:
            evaluations = self.get(
                '/evaluations/dataset/',
                json={
                    'model_id': model_id,
                    'docker_image': docker_image,
                    'pipeline_version': pipeline_version,
                    'hparams': hparams,
                    'dataset_id': dataset_id,
                },
                params={
                    'skip': skip,
                    'limit': api_limit_size,
                }
            )
            for evaluation in evaluations:
                yield evaluation
            if len(evaluations) == api_limit_size:
                skip += api_limit_size
            else:
                break

    @staticmethod
    def _validate_flight(flight: dict) -> dict:
        if (
            (flight['planting_date'] or flight['gdd'] is not None or flight['days_after_planting'] is not None)
            and flight['planting_date_guessed'] is None
        ):
            raise ValueError("Must provide planting_date_guessed if providing planting date")

        if flight['crop_type'] and flight['crop_type_guessed'] is None:
            raise ValueError("Must provide crop_type_guessed if providing crop type")

        return {k: v.isoformat() if isinstance(v, datetime) else v for k, v in flight.items()}

    def create_flight(self, flight_code: str, date: Optional[datetime] = None, provider: Optional[str] = None,
                      planting_date: Optional[datetime] = None, planting_date_guessed: Optional[bool] = None,
                      gdd: Optional[float] = None, days_after_planting: Optional[int] = None,
                      days_after_emergence: Optional[int] = None, growth_stage: Optional[str] = None,
                      crop_type: Optional[str] = None, crop_type_guessed: Optional[bool] = None) -> dict:
        """Creates a flight with the ml service"""
        flight = {
            'flight_code': flight_code,
            'date': date,
            'provider': provider,
            'planting_date': planting_date,
            'planting_date_guessed': planting_date_guessed,
            'gdd': gdd,
            'days_after_planting': days_after_planting,
            'days_after_emergence': days_after_emergence,
            'growth_stage': growth_stage,
            'crop_type': crop_type,
            'crop_type_guessed': crop_type_guessed,
        }
        flight = self._validate_flight(flight)

        return self.post('/flight/', json=flight)

    def get_flight(self, flight_code: str) -> dict:
        """Gets a flight by flight code"""
        return self.get(f'/flight/{flight_code}')

    def update_flight(self, flight: dict) -> dict:
        """Updates a flight in the ml service"""
        flight = self._validate_flight(flight)
        return self.patch(f'/flight/', json=flight)

    def bulk_update_flights(self, flights: List[dict], api_limit_size: int = 100) -> None:
        """Does a bulk update for a list of flights"""
        parsed_flights = list()
        for flight in flights:
            parsed_flights.append(self._validate_flight(flight))

        for i in range(0, len(flights), api_limit_size):
            self.patch('/flight/bulk_update', json=parsed_flights[i:i + api_limit_size])

    def add_flights_to_dataset(self, dataset_id: int, flight_codes: List[str], split: Optional[str] = None,
                               api_limit_size: int = 1000):
        """Adds the given flights codes to the specified dataset with the given split"""
        for i in range(0, len(flight_codes), api_limit_size):
            self.post('/flight/add_to_dataset', json={
                'dataset_id': dataset_id,
                'flight_codes': flight_codes[i:i + api_limit_size],
                'split': split
            })

    def remove_flights_from_dataset(self, dataset_id: int, flight_codes: List[str], api_limit_size: int = 1000):
        """Removes the flight codes for the dataset"""
        for i in range(0, len(flight_codes), api_limit_size):
            self.delete('/flight/remove_from_dataset', json={
                'dataset_id': dataset_id,
                'flight_codes': flight_codes[i:i + api_limit_size]
            })

    def get_flight_datasets(self, flight_code: str, api_limit_size: int = 100) -> Dict[str, List[dict]]:
        """Gets all the datasets that contain this flight"""
        offset = 0
        final_results = defaultdict(list)
        while True:
            results = self.get(f'/flight/datasets',
                               json={
                                   'flight_code': flight_code,
                                   'offset': offset,
                                   'limit': api_limit_size,
                               })
            for split, datasets in results.items():
                final_results[split].extend(datasets)
            if sum([len(datasets) for datasets in results.values()]) == api_limit_size:
                offset += api_limit_size
            else:
                break
        return final_results



import setuptools

with open('src/intelinair_utils/__version__.py') as version_file:
    exec(version_file.read())

with open("requirements.txt", "r") as requirements_file:
    requirements = requirements_file.read().splitlines()

required = [req for req in requirements if not req.startswith('shapely')]

extras = {
    'shapely': ['shapely~=2.0']
}

setuptools.setup(
    name='intelinair_utils',
    version=globals()['__version__'],
    package_dir={'': 'src'},
    packages=setuptools.find_namespace_packages(where='src'),
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=required,
    extras_require=extras
)



import json
import logging
import os
import subprocess
import sys
from typing import Tuple, Union, List

from intelinair_utils.pretty_popen import PrettyPopen

__all__ = ["get_error_message", "call_command", "list_all_files_in_dir"]


def get_logger(name=None, logging_level=logging.INFO):
    _logger = logging.getLogger(name)
    _logger.setLevel(logging_level)

    format_str = "[{asctime}][{levelname}][{name}]: {message}"

    # default stream handler
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(logging.Formatter(format_str, style="{"))

    _logger.addHandler(handler)

    return _logger


logger = get_logger()


def get_error_message(output_folder: str, return_code: int) -> Union[str, None]:
    """Reads the error.json or error.txt file from the
    `output_folder` and returns the content

    Args:
        output_folder (str): folder where is located error.[json, txt] file
        return_code (int): process return code

    Returns:
        Union[str, None]: None if file doesn't exist or file content
    """
    filename = os.path.join(output_folder, 'error.json')
    if os.path.exists(filename):
        try:
            with open(filename) as FL:
                error_data = json.load(FL)
                if "error_code" in error_data and "message" in error_data:
                    return error_data
        except Exception as e:
            if return_code != 0:
                logger.error(f"Exception={e} caught in get_error_message!")
                logger.exception("Error while reading error.json")
    else:
        filename = os.path.join(output_folder, 'error.txt')
        try:
            if not os.path.exists(filename):
                return None
            with open(filename) as f:
                error_code = f.readline()[:-1]
                error_message = f.readline().strip()
                return error_code + ": " + error_message
        except Exception as e:
            if return_code != 0:
                logger.error(f"Exception={e} caught in get_error_message!")
                logger.exception("Error while reading error.txt")
    return


def call_command(cmd: str, command_name: str, output_folder: str = None,
                 silent: bool = False, print_command: bool = True) -> Tuple[int, str, str]:
    """Calls a shell command, prints stdout/stderr and returns the return code
    Known issue: stdout and stderr are not synced
    Also tries to read error.txt from output_folder and return it
    Finally, stderr is also returned

    Args:
        cmd (str): command (like in Unix shell)
        command_name (str): name of the command (for logging)
        output_folder (str, optional): The output folder where the command 
                                       will save error.[json,txt] outpur. Defaults to None.
        silent (bool, optional): Prints the command output on failure. Defaults to False.
        print_command (bool, optional): if true will print the command. Defaults to True.

    Returns:
        [Tuple]: exit code of the command, error message from error.txt, stderr output
    """
    if not silent and print_command:
        logger.info("Calling {}".format(cmd))

    # TODO: timeout?
    process = PrettyPopen(args=cmd,
                          shell=True,
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE)

    out, err = process.communicate()

    if out and not silent:
        logger.info(out.decode('utf-8', 'ignore'))
    if err:
        logger.info("Subprocess stderr:")
        logger.info(err.decode('utf-8', 'ignore'))

    if process.returncode != 0 or not silent:
        logger.info("Command {} is completed with error code {}".format(
            command_name, process.returncode))

    if output_folder is not None:
        error_msg_from_txt = get_error_message(output_folder, process.returncode)
        return process.returncode, error_msg_from_txt, err.decode('utf-8', 'ignore')
    error_message = None
    if process.returncode:
        error_message = f"Return code: {process.returncode}"

    return process.returncode, error_message, err.decode('utf-8', 'ignore')


def list_all_files_in_dir(dir_path: str, recursive: bool = True, search_str: str = "",
                          suffix_match: bool = False) -> List[str]:
    """Lists all files in a directory

    Args:
        dir_path (str): path to the directory
        recursive (bool, optional): if true will list recursively. Defaults to True.
        search_str (str, optional): if provided will only return files containing this string.
        suffix_match (bool, optional): If true, will match `search_str` as a suffix in filenames only.
        Defaults to False.


    Returns:
        List[str]: List of files matching the specified criteria.
    """
    if not os.path.exists(dir_path):
        return []
    file_list = []
    if recursive:
        for path, _, files in os.walk(dir_path):
            for file_name in files:
                full_path = os.path.join(path, file_name)
                if suffix_match:
                    if file_name.endswith(search_str):
                        file_list.append(full_path)
                elif search_str in file_name:
                    file_list.append(full_path)
    else:
        for file_name in os.listdir(dir_path):
            full_path = os.path.join(dir_path, file_name)
            if os.path.isfile(full_path):
                if suffix_match:
                    if file_name.endswith(search_str):
                        file_list.append(full_path)
                elif search_str in file_name:
                    file_list.append(full_path)

    return file_list



import logging
import os
from pathlib import Path
from typing import Union, Dict, Optional


class ErrorLogger:
    """Special class for logging error codes in image processing"""

    def __init__(self, error_codes: Dict[int, str], output_dir: Union[str, Path] = None, logger: logging.Logger = None):
        self.logger = logger
        self.output_dir = output_dir
        self.error_codes = error_codes

        self.error_file_name = "error.txt"
        self.recorded_errors = []

    def log_error_with_code(self, code: int, info: Optional[str] = None):
        """Log an error with a given error code"""
        if self.output_dir is None:
            raise Exception("output_dir is None.")

        log_msg = f"recording error with code: {code} ({self.error_codes[code]})"
        if info:
            log_msg += f" - info: {info}"

        if self.logger:
            if str(code).startswith("2"):
                self.logger.warning(log_msg)
            else:
                self.logger.error(log_msg)

        self.recorded_errors.append(code)

        error_path = os.path.join(self.output_dir, self.error_file_name)
        error_message = self.error_codes[code]
        if info:
            error_message += f" - info: {info}"

        with open(error_path, "a") as f:
            f.write(str(code) + "\n")
            f.write(error_message + "\n")

    def log_unknown_error_with_message(self, error_message: str, info: Optional[str] = None):
        """Log an error with a given error code"""
        if self.output_dir is None:
            raise Exception("output_dir is None.")

        self.logger.error(error_message)

        self.recorded_errors.append(3)

        error_path = os.path.join(self.output_dir, self.error_file_name)
        if info:
            error_message += f" - info: {info}"

        with open(error_path, "a") as f:
            f.write('error without code logged' + "\n")
            f.write(error_message + "\n")

    def last_recorded_error(self):
        """Returns last recorded error code. Returns 0 if no codes have been previously recorded."""
        if len(self.recorded_errors):
            return self.recorded_errors[-1]

        return 0



"""
Helper utilities for api.
"""
from datetime import datetime


def validate_date(date_to_validate: str):
    """
    Date string validation method.
    Currently supported formats are:
    '%Y-%m-%d'              -> 2021-10-27
    '%Y-%m-%dT%H:%M:%SZ'    -> 2022-08-13T10:37:23Z
    Args:
        date_to_validate: input date string to validate
    Raises:
        ValueError: In case of invalid input date format
    """
    valid_date_formats = ['%Y-%m-%d', '%Y-%m-%dT%H:%M:%SZ']
    for date_format in valid_date_formats:
        try:
            datetime.strptime(date_to_validate, date_format)
        except ValueError:
            continue
        break
    else:
        raise ValueError(f'Invalid date string:"{date_to_validate}" valid formats are: {", ".join(valid_date_formats)}')



import json


def compare_(obj1, obj2):
    if type(obj1) != type(obj2):
        return False
    if isinstance(obj1, list):
        if len(obj2) != len(obj1):
            return False
        for i in range(len(obj1)):
            if not compare_(obj1[i], obj2[i]):
                return False
    if isinstance(obj1, dict):
        if len(obj1) != len(obj2):
            return False
        for key, value in obj1.items():
            if key not in obj2:
                return False
            if not compare_(value, obj2[key]):
                return False
    return obj1 == obj2


def compare_jsons(fp1, fp2):
    with open(fp1) as f1:
        o1 = json.load(f1)
    with open(fp2) as f2:
        o2 = json.load(f2)
    return compare_(o1, o2)



"""Authorization token storage and retrieval classes for AgMRI API
"""
import json
import logging
import os.path
import string
from collections import namedtuple

import boto3
import botocore.exceptions
import filelock

logger = logging.getLogger(__name__)

__all__ = ['SSMStore', 'FileStore', 'TokenPair']

PROCESSING_TOKENS_TEMPLATE = "/global/processing/tokens/{environment}/{username}/{token_type}"

"""Authorization tokens stored together

Attributes:
    access_token: JWT access token with a short lifetime
    refresh_token: Refresh token with a long lifetime that can
                   be used to get a new access token
"""
TokenPair = namedtuple('TokenPair', ['access_token', 'refresh_token'])


class SSMStore:
    """Store and retrieve tokens from the AWS SSM Parameter Store
    """

    def __init__(self, environment, username):
        self.environment = environment
        self.username = self.clean_username(username)

    @staticmethod
    def clean_username(username):
        """Remove all non-alphanumeric characters from the username

        Args:
            username: the username (usually an email) to clean

        Returns:
            the cleaned username
        """
        return ''.join([c if c in string.ascii_letters + string.digits else '_' for c in username])

    def get_tokens(self) -> TokenPair:
        """Retrieve the tokens from the parameter store
        """
        # Not defining as a property as ssm client is not picklable which is an issue for cropping metaflow pipeline
        ssm = boto3.client('ssm')
        token_dict = {}
        for token_type in TokenPair._fields:
            name = PROCESSING_TOKENS_TEMPLATE.format(environment=self.environment,
                                                     username=self.username,
                                                     token_type=token_type)
            parameter = ssm.get_parameter(Name=name, WithDecryption=True)
            token_dict[token_type] = parameter['Parameter']['Value']
        return TokenPair(**token_dict)

    def _store_token(self, token_type: str, token_value: str):
        """Store a token in the parameter store

        Args:
            token_type: the type of token to store
            token_value: the value of the token
        """
        # Not defining as a property as ssm client is not picklable which is an issue for cropping metaflow pipeline
        ssm = boto3.client('ssm')
        try:
            name = PROCESSING_TOKENS_TEMPLATE.format(environment=self.environment,
                                                     username=self.username,
                                                     token_type=token_type)
            ssm.put_parameter(
                Name=name,
                Value=token_value,
                Type='SecureString',
                Overwrite=True
            )
        except botocore.exceptions.ClientError as exc:
            # Boto3 classifies all AWS service errors and exceptions as ClientError exceptions
            logger.warning(
                f"Failed to store token to the parameter store with exception {exc.response['Error']['Code']}"
            )

    def store_tokens(self, tokens: TokenPair):
        """Store the tokens in the parameter store

        Args:
            tokens: the tokens to store
        """
        for token_type, token_value in tokens._asdict().items():
            self._store_token(token_type, token_value)


class FileStore:
    """Store and retrieve tokens from a local file
    """

    def __init__(self, environment, username, path=None):
        self.environment = environment
        self.username = username
        self.token_path = path if path is not None else os.path.expanduser(f'~/.agmri.{environment}.tokens')
        self.token_lock_path = self.token_path + '.lock'

    def get_tokens(self):
        """Retrieve the tokens from the file
        """
        with filelock.FileLock(self.token_lock_path):
            with open(self.token_path, 'r', encoding='utf-8') as tokens_file:
                tokens_json = json.load(tokens_file)
                return TokenPair(**tokens_json)

    def store_tokens(self, tokens: TokenPair):
        """Store the tokens in the file

        Args:
            tokens: the tokens to store
        """
        with filelock.FileLock(self.token_lock_path):
            with open(self.token_path, 'w', encoding='utf-8') as tokens_file:
                json.dump(tokens._asdict(), tokens_file)



import json
import os
from logging import getLogger
from typing import Optional, Union, Dict, Callable

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from intelinair_utils.agmri_api import AgmriApi
from intelinair_utils.api import Api

logger = getLogger(__name__)

URL = "https://api.{}.intelinair.dev/alerts/api/"
PROD_URL = "https://api.ag-mri.intelinair.com/alerts/api/"
REQUEST_TIMEOUT = 180


class AlertServiceApi(Api):
    """An API for getting alerts from Alert Service."""

    def __init__(self, environment: str, agmri_api: AgmriApi, version_string: str = 'v1/'):
        # version_string == '' corresponds to the first version of API
        self.environment = environment
        self.version_string = version_string
        self.agmri_api = agmri_api
        url = PROD_URL if self.environment == 'prod' else URL.format(self.environment)
        self.url = os.path.join(url, self.version_string)
        self.auth_header = {
            'Content-Type': 'application/json',
            **self.agmri_api.headers["processing"]
        }
        self._session = requests.Session()
        retries = Retry(total=10, read=10, connect=10, backoff_factor=1, allowed_methods=False,
                        status_forcelist=[409] + list(range(500, 600)))
        self._session.mount('https://', HTTPAdapter(max_retries=retries))

    def _make_request(self, method: Callable, path: str, body: Optional[str] = None) -> Dict:
        """
        Makes a request to the Alert Service API of the given method.
        Args:
            method: The method to use for the request.
            path: The path to the endpoint.
            body: The body of the request if POST.
        Returns:
            The response of the request.
        """
        res = method(os.path.join(self.url, path), headers=self.auth_header, timeout=REQUEST_TIMEOUT, data=body)
        if res.status_code == 401:
            self.agmri_api.ensure_authenticated()
            self.auth_header.update(self.agmri_api.headers["processing"])
            res = method(os.path.join(self.url, path), headers=self.auth_header, timeout=REQUEST_TIMEOUT, data=body)
        if res.status_code == 204:
            return dict()
        return res.json()

    def get(self, path: str):
        """
        Performs a GET request to the specified endpoint.
        Args:
            path: The path to the endpoint.
        Returns:
            The json response of the request.
        """
        return self._make_request(self._session.get, path)

    def post(self, path, body: Optional[Union[Dict, str]] = None):
        """
        Performs a POST request to the specified endpoint.
        Args:
            path: The path to the endpoint.
            body: The body of the request.
        Returns:
            The json response of the request.
        """
        if isinstance(body, dict):
            body = json.dumps(body)
        return self._make_request(self._session.post, path, body)

    def delete(self, path: str):
        """
        Performs a DELETE request to the specified endpoint.
        Args:
            path: The path to the endpoint.
        Returns:
            The json response of the request.
        """
        return self._make_request(self._session.delete, path)
